//CLEO:Android Extender v1.0
//
//by 3LC, made from Fastman92's sample plugin + FLA source code

#include "dllmain.h"
// from $(FASTMAN92_INCLUDE) directory
#include <Assembly/eInstructionSet.h>
#include <Assembly/UsefulMacros.h>
#include <MemoryAddressCalculator/CMemoryAddressCalculator.h>
#include <MemoryPermission/MemoryPermission.h>
// from 'fastman92 plugin loader'
#include "../../fastman92 limit adjuster/fastman92 plugin loader/ForOtherProjects/Exports.h"
////////
#include <android/log.h>
#include <limits.h>
#include <dlfcn.h>
// Added for CleoExtender
#include <fileIO.cpp> //for pluginSize
#include <stdio.h> //for sprintf
#include <dirent.h> //for dirent

static void OutputFormattedDebugStringVargs(const char* format, va_list args)
{
	__android_log_vprint(ANDROID_LOG_INFO, "CleoExtender", format, args);
}

void OutputFormattedDebugString(const char* format, ...)
{
	va_list args;
	va_start(args, format);
	OutputFormattedDebugStringVargs(format, args);
}

// Set of functions used to write to protected memory
class CPatch
{
public:
	static void WriteDataToMemory(void* dwAddress, const void* bData, int iSize)
	{
		tMemoryPermissionChangeRequest request;
		request.lpAddress = dwAddress;
		request.dwSize = iSize;
		request.flNewProtect = GetNativeNewProtect(F92_MEM_PAGE_EXECUTE_READWRITE);

		SetMemoryPermission(&request);

		memcpy(dwAddress, bData, iSize);

		// Restore old permission
		if (request.bIsOldProtectSet)
		{
			request.flNewProtect = request.lpflOldProtect;
			SetMemoryPermission(&request);
		}
	}

	// Patches memory data
	static void PatchMemoryData(uintptr_t dwAddress, const void* bData, int iSize)
	{
		WriteDataToMemory((void*)dwAddress, bData, iSize);
	}

	// Patches BYTE value
	static void PatchBYTE(uintptr_t dwAddress, uint8_t to)
	{
		WriteDataToMemory((void*)dwAddress, &to, sizeof(to));
	}

	// Patches WORD value
	static void PatchWORD(uintptr_t dwAddress, uint16_t to)
	{
		WriteDataToMemory((void*)dwAddress, &to, sizeof(to));
	}

	// Patches DWORD value
	static void PatchDWORD(uintptr_t dwAddress, uint32_t to)
	{
		WriteDataToMemory((void*)dwAddress, &to, sizeof(to));
	}

	// Patches pointer
	void PatchPointer(uintptr_t dwAddress, const void* to)
	{
		WriteDataToMemory((void*)dwAddress, &to, sizeof(to));
	}

	// Redirects the code, see the comments
	static void RedirectCodeEx(
		eInstructionSet sourceInstructionSet,
		uintptr_t dwAddress,
		const void* to)
	{
		// Thumb trampoline may take 8 bytes (if address is aligned to value of 4) or 10 bytes.
		if (sourceInstructionSet == INSTRUCTION_SET_THUMB)
		{
			char code[12];
			unsigned int sizeOfData = 0;

			if (dwAddress % 4 == 0)
			{
				*(uint32_t*)(code + 0) = 0xF000F8DF;	// // LDR.W             R0, [PC, #0]
				*(const void**)(code + 4) = to;	// pointer, where to jump
				sizeOfData = 8;
			}
			else
			{
				*(uint32_t*)(code + 0) = 0xBF00;	// NOP
				*(uint32_t*)(code + 2) = 0xF000F8DF;	// // LDR.W             R0, [PC, #0]

				*(const void**)(code + 6) = to;	// pointer, where to jump
				sizeOfData = 10;
			}

			CPatch::PatchMemoryData(dwAddress, code, sizeOfData);
		}
		// ARM trampoline takes 8 bytes
		else if (sourceInstructionSet == INSTRUCTION_SET_ARM)
		{
			char code[8];

			*(uint32_t*)(code + 0) = 0xE51FF004;	// // LDR             R0, [PC, #0]
			*(const void**)(code + 4) = to;	// pointer, where to jump
			CPatch::PatchMemoryData(dwAddress, code, sizeof(code));
		}
	}

	void RedirectCodeEx(
		eInstructionSet sourceInstructionSet,
		uintptr_t dwAddress,
		uintptr_t to)
	{
		CPatch::RedirectCodeEx(sourceInstructionSet, dwAddress, (const void*)to);
	}

	// Redirects method
	template<typename T>static void RedirectMethodEx(eInstructionSet instructionSet, uintptr_t dwAddress, T method)
	{
		auto _ptr = method;
		void*& ptr = (void*&)_ptr;
		RedirectCodeEx(instructionSet, dwAddress, ptr);
	}
};

void GetApplicationLibDirectoryPath(char* libPluginPath)
{
	Dl_info info;
	int info_result = dladdr((const void*)&GetApplicationLibDirectoryPath, &info);

	// remove a filename
	int index_of_last_slash = 0;

	for (unsigned int i = 0; info.dli_fname[i]; i++)
	{
		char c = info.dli_fname[i];

		if (c == '/' || c == '\\')
			index_of_last_slash = i;
	}

	memcpy(libPluginPath, info.dli_fname, (index_of_last_slash));
	libPluginPath[index_of_last_slash] = 0;
}

void GetCleoScriptsDirectoryPath(char* CleoScriptsDirectory)
{
	const char* rootDirectory = GetStorageRootDirectoryPath();
	char* rootDir;
	strcpy(rootDir, rootDirectory); 
	
	// remove subfolders
	char* FilesFolder;
	FilesFolder = strstr(rootDir, "files/");

	int pathLength = (strlen(rootDirectory) - strlen(FilesFolder));

	memcpy(CleoScriptsDirectory, rootDirectory, pathLength);
	CleoScriptsDirectory[pathLength] = 0;
}

// ASLR handling
CMemoryAddressCalculator g_mCalc;	// address calculator
JavaVM* g_javaVM;	// Java virtual machine
void* g_applicationLibHandle;
bool CustomTextExists = false;

extern "C"
{
	bool DLL_EXPORT OnPluginLoad(
		tPluginStartParams* pPluginStartParams
	)
	{
		g_javaVM = pPluginStartParams->jvm;
		g_applicationLibHandle = pPluginStartParams->applicationLibHandle;
		g_mCalc.Initialize(0, (uintptr_t)pPluginStartParams->applicationLibBaseAddress);
		OutputFormattedDebugString(
			"OnPluginLoad. Current application library base = 0x%llX",
			(uint64_t)g_mCalc.GetCurrentModuleBase()
		);
		return true;
	}

	bool DLL_EXPORT OnApplicationInit()
	{
		static char libPluginPath[PATH_MAX] = "";
		static char CleoScriptsDirectory[PATH_MAX] = "";
		char CleoPath[PATH_MAX]; 
		GetApplicationLibDirectoryPath(libPluginPath);
		GetCleoScriptsDirectoryPath(CleoScriptsDirectory);
		sprintf(CleoPath, "%s/libcleo.so", libPluginPath);

		void* CleoHandle = dlopen(CleoPath, RTLD_GLOBAL | RTLD_LAZY);
		if (CleoHandle)
		{
			OutputFormattedDebugString("Found Cleo at %s", CleoPath);

			auto CleoSize = GetFileSize(CleoPath);

			Dl_info info;
			const char* ver_str;

			uintptr_t stub_jni_OnLoad = (uintptr_t)dlsym(CleoHandle, "stub_JNI_OnLoad");

			if (dladdr((const void*)stub_jni_OnLoad, &info))
			{
				g_mCalc.Initialize(0, (uintptr_t)info.dli_fbase);	// set image base

				uintptr_t stub_jni_OnLoad_preferred = g_mCalc.GetPreferedVAbyCurrentVA(stub_jni_OnLoad);

				if (CleoSize == 83232
					&& GET_CODE_START(stub_jni_OnLoad_preferred) == 0x8084) //CLEO Android 2016
				{
					ver_str = (const char*)g_mCalc.GetCurrentVAbyPreferedVA(0x122B4);

					if (!strcmp(ver_str, "CLEO ANDROID (04 May 2016)"))
					{
						////CLEO MENU EDITS
						//The Cleo Menu is an internal scm/cleo script ('menu'). These lines directly edit the script in hex.

						{	////Disable "Player.Controllable" check
							//Allows opening menu while swimming/falling/etc. Changes 83EE opcodes to 8256.
							CPatch::PatchWORD(g_mCalc.GetCurrentVAbyPreferedVA(0x10428), 0x8256);
							CPatch::PatchWORD(g_mCalc.GetCurrentVAbyPreferedVA(0x10C08), 0x8256);
							CPatch::PatchWORD(g_mCalc.GetCurrentVAbyPreferedVA(0x11405), 0x8256);
							////Disable camera cuts
							//Stops resetting camera upon open/close of menu. Changes 02EB & 0373 opcodes to nop.
							CPatch::PatchWORD(g_mCalc.GetCurrentVAbyPreferedVA(0x106EE), 0x0000); //02EB
							CPatch::PatchWORD(g_mCalc.GetCurrentVAbyPreferedVA(0x10ED0), 0x0000);
							CPatch::PatchWORD(g_mCalc.GetCurrentVAbyPreferedVA(0x11487), 0x0000);
							CPatch::PatchWORD(g_mCalc.GetCurrentVAbyPreferedVA(0x104A6), 0x0000); //0373
							CPatch::PatchWORD(g_mCalc.GetCurrentVAbyPreferedVA(0x106F0), 0x0000);
							CPatch::PatchWORD(g_mCalc.GetCurrentVAbyPreferedVA(0x10C88), 0x0000);
							CPatch::PatchWORD(g_mCalc.GetCurrentVAbyPreferedVA(0x10ED2), 0x0000);
							CPatch::PatchWORD(g_mCalc.GetCurrentVAbyPreferedVA(0x11489), 0x0000);
							OutputFormattedDebugString("Cleo Menu Unlocked (Anim checks & camera resets disabled).");
						}

						////CUSTOM CLEO TEXT
						//Erase internal GXT titles, if CustomCleoText.fxt is found

						DIR* dir = opendir(CleoScriptsDirectory);
						dirent* drnt;

						while ((drnt = readdir(dir)) != nullptr)
						{
							if ((!strcmp(drnt->d_name, "CustomCleoText.fxt")) || (!strcmp(drnt->d_name, "customcleotext.fxt")))
							{
								CustomTextExists = true;
							}
							continue;
						}
						closedir(dir);

						if (CustomTextExists)
						{
							CPatch::WriteDataToMemory((void*)g_mCalc.GetCurrentVAbyPreferedVA(0x125F8), "\x00\x00\x00\x00\x00\x00", 6); //SPLASH
							CPatch::WriteDataToMemory((void*)g_mCalc.GetCurrentVAbyPreferedVA(0x1261C), "\x00\x00\x00\x00\x00", 5);     //CLDSC
							CPatch::WriteDataToMemory((void*)g_mCalc.GetCurrentVAbyPreferedVA(0x12668), "\x00\x00\x00\x00\x00\x00", 6); //CLDSCL
							CPatch::WriteDataToMemory((void*)g_mCalc.GetCurrentVAbyPreferedVA(0x126B0), "\x00\x00\x00\x00\x00", 5);     //CLMNU
							CPatch::WriteDataToMemory((void*)g_mCalc.GetCurrentVAbyPreferedVA(0x126C4), "\x00\x00\x00\x00\x00\x00", 6); //CLMNUD
							CPatch::WriteDataToMemory((void*)g_mCalc.GetCurrentVAbyPreferedVA(0x1270C), "\x00\x00\x00\x00\x00\x00", 6); //CLMNU0
							CPatch::WriteDataToMemory((void*)g_mCalc.GetCurrentVAbyPreferedVA(0x12728), "\x00\x00\x00\x00\x00\x00", 6); //CLMNU1
							CPatch::WriteDataToMemory((void*)g_mCalc.GetCurrentVAbyPreferedVA(0x1274C), "\x00\x00\x00\x00\x00\x00", 6); //CLMNU2
							CPatch::WriteDataToMemory((void*)g_mCalc.GetCurrentVAbyPreferedVA(0x127C0), "\x00\x00\x00\x00\x00\x00", 6); //CLMNUL
							CPatch::WriteDataToMemory((void*)g_mCalc.GetCurrentVAbyPreferedVA(0x127D0), "\x00\x00\x00\x00\x00\x00", 6); //CLMNUT
							CPatch::WriteDataToMemory((void*)g_mCalc.GetCurrentVAbyPreferedVA(0x127DC), "\x00\x00\x00\x00\x00\x00", 6); //CLMNUB
							CPatch::WriteDataToMemory((void*)g_mCalc.GetCurrentVAbyPreferedVA(0x127EC), "\x00\x00\x00\x00\x00\x00", 6); //CLMNUR
							CPatch::WriteDataToMemory((void*)g_mCalc.GetCurrentVAbyPreferedVA(0x127FC), "\x00\x00\x00\x00\x00\x00", 6); //CLMNUS
							OutputFormattedDebugString("Found CustomCleoText.fxt - Cleo Text Unlocked.");
						}
						else OutputFormattedDebugString("CustomCleoText.fxt not found in %s - Leaving default text.", CleoScriptsDirectory);

					}
					else OutputFormattedDebugString("Incorrect Cleo version - must be May4-2016.");
				}
				else OutputFormattedDebugString("Incorrect Cleo version - must be May4-2016.");
			}
			else OutputFormattedDebugString("Cleo hooking failed for unknown reasons.");
		}
		else OutputFormattedDebugString("Cleo not found at %s - Nothing patched.", CleoPath);
		
		return true;
	}
}